"""
A mock client for testing functions that call the minecraft server.
This mocks a world of 100 blocks square with origin in the middle.
"""

from mcwb.types import Items, Vec3
import numpy as np

from mcipc.rcon.enumerations import Item


class MockClient:
    def __init__(self):
        """ create an empty mock world """
        self.world = np.full((100, 100, 100), Item.AIR, dtype=Item)
        self.offset = Vec3(50, 50, 50)

    # the following are mock versions of the original Client Functions

    def setblock(self, position: Vec3, block: Item):
        """ set the block at position in the world """
        pos = position - self.offset
        self.world[pos.x, pos.y, pos.z] = block

    def fill(self, start: Vec3, end: Vec3, block: Item):
        block = Item(block)  # ensure enum
        start -= self.offset
        end -= self.offset - 1
        self.world[start.x:end.x, start.y:end.y, start.z:end.z] = block

    # the following are additional functions for use in tests

    def getblock(self, position: Vec3) -> Item:
        """ return the block at position in the world """
        pos = position - self.offset
        return self.world[pos.x, pos.y, pos.z]

    def compare(self, position: Vec3, cube: Items):
        """ verify that the contents of the world at pos matches cube """
        ncube = np.array(cube)
        pos = position - self.offset
        upper = pos + Vec3(*(ncube.shape))
        world_slice = self.world[pos.x:upper.x, pos.y:upper.y, pos.z:upper.z]

        return np.array_equal(world_slice, ncube)



"""Tests the functions module."""

from functools import partial
from itertools import product
from unittest import TestCase

from mcipc.rcon.enumerations import Item
from mcwb.functions import get_direction, normalize, offsets, validate
from mcwb.types import Anchor, Direction, Vec3


class TestGetDirection(TestCase):
    """Tests the get_direction() function."""

    def setUp(self):
        """Sets example directions."""
        self.valid_direction_x = (Vec3(10, -9, 3), Vec3(91, -9, 3))
        self.direction_vector_x = Vec3(81, 0, 0)
        self.valid_direction_y = (Vec3(10, -9, 3), Vec3(10, 200, 3))
        self.direction_vector_y = Vec3(0, 209, 0)
        self.valid_direction_z = (Vec3(10, -9, 3), Vec3(10, -9, -5))
        self.direction_vector_z = Vec3(0, 0, -8)
        self.invalid_direction = (Vec3(1, 2, -3), Vec3(-4, -5, 6))

    def test_valid_direction_x(self):
        """Tests the results of get_direction()."""
        self.assertEqual(get_direction(*self.valid_direction_x),
                         self.direction_vector_x)
        self.assertEqual(get_direction(*self.valid_direction_y),
                         self.direction_vector_y)
        self.assertEqual(get_direction(*self.valid_direction_z),
                         self.direction_vector_z)
        self.assertRaises(ValueError,
                          partial(get_direction, *self.invalid_direction))


class TestNormalize(TestCase):
    """Tests the normalize() function."""

    def setUp(self):
        """Sets example profiles."""
        self.valid_profile_template = [
            [Item.RED_CONCRETE, None, Item.GREEN_CONCRETE],
            [None, None, None],
            [Item.BLUE_CONCRETE, None, Item.YELLOW_CONCRETE],
        ]
        self.valid_profile_air = [
            [Item.RED_CONCRETE, Item.AIR, Item.GREEN_CONCRETE],
            [Item.AIR, Item.AIR, Item.AIR],
            [Item.BLUE_CONCRETE, Item.AIR, Item.YELLOW_CONCRETE],
        ]

    def test_normalize(self):
        """Tests the normalization of the profile template."""
        self.assertEqual(
            list(normalize(self.valid_profile_template, Item.AIR)),
            self.valid_profile_air)


class TestOffsets(TestCase):
    """Tests the offsets() function."""

    def setUp(self):
        """Sets example profiles, directions and vectors."""
        self.valid_profile = [
            [Item.RED_CONCRETE, Item.AIR, Item.GREEN_CONCRETE],
            [Item.AIR, Item.AIR, Item.AIR],
            [Item.BLUE_CONCRETE, Item.AIR, Item.YELLOW_CONCRETE],
        ]
        self.directions = Direction
        self.anchors = Anchor
        self.results = [
            [
                (Item.RED_CONCRETE, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=0, y=0, z=1)),
                (Item.GREEN_CONCRETE, Vec3(x=0, y=0, z=2)),
                (Item.AIR, Vec3(x=0, y=-1, z=0)),
                (Item.AIR, Vec3(x=0, y=-1, z=1)),
                (Item.AIR, Vec3(x=0, y=-1, z=2)),
                (Item.BLUE_CONCRETE, Vec3(x=0, y=-2, z=0)),
                (Item.AIR, Vec3(x=0, y=-2, z=1)),
                (Item.YELLOW_CONCRETE, Vec3(x=0, y=-2, z=2))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=0, y=0, z=-2)),
                (Item.AIR, Vec3(x=0, y=0, z=-1)),
                (Item.GREEN_CONCRETE, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=0, y=-1, z=-2)),
                (Item.AIR, Vec3(x=0, y=-1, z=-1)),
                (Item.AIR, Vec3(x=0, y=-1, z=0)),
                (Item.BLUE_CONCRETE, Vec3(x=0, y=-2, z=-2)),
                (Item.AIR, Vec3(x=0, y=-2, z=-1)),
                (Item.YELLOW_CONCRETE, Vec3(x=0, y=-2, z=0))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=0, y=2, z=0)),
                (Item.AIR, Vec3(x=0, y=2, z=1)),
                (Item.GREEN_CONCRETE, Vec3(x=0, y=2, z=2)),
                (Item.AIR, Vec3(x=0, y=1, z=0)),
                (Item.AIR, Vec3(x=0, y=1, z=1)),
                (Item.AIR, Vec3(x=0, y=1, z=2)),
                (Item.BLUE_CONCRETE, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=0, y=0, z=1)),
                (Item.YELLOW_CONCRETE, Vec3(x=0, y=0, z=2))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=0, y=2, z=-2)),
                (Item.AIR, Vec3(x=0, y=2, z=-1)),
                (Item.GREEN_CONCRETE, Vec3(x=0, y=2, z=0)),
                (Item.AIR, Vec3(x=0, y=1, z=-2)),
                (Item.AIR, Vec3(x=0, y=1, z=-1)),
                (Item.AIR, Vec3(x=0, y=1, z=0)),
                (Item.BLUE_CONCRETE, Vec3(x=0, y=0, z=-2)),
                (Item.AIR, Vec3(x=0, y=0, z=-1)),
                (Item.YELLOW_CONCRETE, Vec3(x=0, y=0, z=0))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=0, y=1, z=-1)),
                (Item.AIR, Vec3(x=0, y=1, z=0)),
                (Item.GREEN_CONCRETE, Vec3(x=0, y=1, z=1)),
                (Item.AIR, Vec3(x=0, y=0, z=-1)),
                (Item.AIR, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=0, y=0, z=1)),
                (Item.BLUE_CONCRETE, Vec3(x=0, y=-1, z=-1)),
                (Item.AIR, Vec3(x=0, y=-1, z=0)),
                (Item.YELLOW_CONCRETE, Vec3(x=0, y=-1, z=1))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=0, y=0, z=-1)),
                (Item.GREEN_CONCRETE, Vec3(x=0, y=0, z=-2)),
                (Item.AIR, Vec3(x=0, y=-1, z=0)),
                (Item.AIR, Vec3(x=0, y=-1, z=-1)),
                (Item.AIR, Vec3(x=0, y=-1, z=-2)),
                (Item.BLUE_CONCRETE, Vec3(x=0, y=-2, z=0)),
                (Item.AIR, Vec3(x=0, y=-2, z=-1)),
                (Item.YELLOW_CONCRETE, Vec3(x=0, y=-2, z=-2))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=0, y=0, z=2)),
                (Item.AIR, Vec3(x=0, y=0, z=1)),
                (Item.GREEN_CONCRETE, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=0, y=-1, z=2)),
                (Item.AIR, Vec3(x=0, y=-1, z=1)),
                (Item.AIR, Vec3(x=0, y=-1, z=0)),
                (Item.BLUE_CONCRETE, Vec3(x=0, y=-2, z=2)),
                (Item.AIR, Vec3(x=0, y=-2, z=1)),
                (Item.YELLOW_CONCRETE, Vec3(x=0, y=-2, z=0))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=0, y=2, z=0)),
                (Item.AIR, Vec3(x=0, y=2, z=-1)),
                (Item.GREEN_CONCRETE, Vec3(x=0, y=2, z=-2)),
                (Item.AIR, Vec3(x=0, y=1, z=0)),
                (Item.AIR, Vec3(x=0, y=1, z=-1)),
                (Item.AIR, Vec3(x=0, y=1, z=-2)),
                (Item.BLUE_CONCRETE, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=0, y=0, z=-1)),
                (Item.YELLOW_CONCRETE, Vec3(x=0, y=0, z=-2))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=0, y=2, z=2)),
                (Item.AIR, Vec3(x=0, y=2, z=1)),
                (Item.GREEN_CONCRETE, Vec3(x=0, y=2, z=0)),
                (Item.AIR, Vec3(x=0, y=1, z=2)),
                (Item.AIR, Vec3(x=0, y=1, z=1)),
                (Item.AIR, Vec3(x=0, y=1, z=0)),
                (Item.BLUE_CONCRETE, Vec3(x=0, y=0, z=2)),
                (Item.AIR, Vec3(x=0, y=0, z=1)),
                (Item.YELLOW_CONCRETE, Vec3(x=0, y=0, z=0))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=0, y=1, z=1)),
                (Item.AIR, Vec3(x=0, y=1, z=0)),
                (Item.GREEN_CONCRETE, Vec3(x=0, y=1, z=-1)),
                (Item.AIR, Vec3(x=0, y=0, z=1)),
                (Item.AIR, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=0, y=0, z=-1)),
                (Item.BLUE_CONCRETE, Vec3(x=0, y=-1, z=1)),
                (Item.AIR, Vec3(x=0, y=-1, z=0)),
                (Item.YELLOW_CONCRETE, Vec3(x=0, y=-1, z=-1))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=1, y=0, z=0)),
                (Item.GREEN_CONCRETE, Vec3(x=2, y=0, z=0)),
                (Item.AIR, Vec3(x=0, y=0, z=-1)),
                (Item.AIR, Vec3(x=1, y=0, z=-1)),
                (Item.AIR, Vec3(x=2, y=0, z=-1)),
                (Item.BLUE_CONCRETE, Vec3(x=0, y=0, z=-2)),
                (Item.AIR, Vec3(x=1, y=0, z=-2)),
                (Item.YELLOW_CONCRETE, Vec3(x=2, y=0, z=-2))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=-2, y=0, z=0)),
                (Item.AIR, Vec3(x=-1, y=0, z=0)),
                (Item.GREEN_CONCRETE, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=-2, y=0, z=-1)),
                (Item.AIR, Vec3(x=-1, y=0, z=-1)),
                (Item.AIR, Vec3(x=0, y=0, z=-1)),
                (Item.BLUE_CONCRETE, Vec3(x=-2, y=0, z=-2)),
                (Item.AIR, Vec3(x=-1, y=0, z=-2)),
                (Item.YELLOW_CONCRETE, Vec3(x=0, y=0, z=-2))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=0, y=0, z=2)),
                (Item.AIR, Vec3(x=1, y=0, z=2)),
                (Item.GREEN_CONCRETE, Vec3(x=2, y=0, z=2)),
                (Item.AIR, Vec3(x=0, y=0, z=1)),
                (Item.AIR, Vec3(x=1, y=0, z=1)),
                (Item.AIR, Vec3(x=2, y=0, z=1)),
                (Item.BLUE_CONCRETE, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=1, y=0, z=0)),
                (Item.YELLOW_CONCRETE, Vec3(x=2, y=0, z=0))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=-2, y=0, z=2)),
                (Item.AIR, Vec3(x=-1, y=0, z=2)),
                (Item.GREEN_CONCRETE, Vec3(x=0, y=0, z=2)),
                (Item.AIR, Vec3(x=-2, y=0, z=1)),
                (Item.AIR, Vec3(x=-1, y=0, z=1)),
                (Item.AIR, Vec3(x=0, y=0, z=1)),
                (Item.BLUE_CONCRETE, Vec3(x=-2, y=0, z=0)),
                (Item.AIR, Vec3(x=-1, y=0, z=0)),
                (Item.YELLOW_CONCRETE, Vec3(x=0, y=0, z=0))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=-1, y=0, z=1)),
                (Item.AIR, Vec3(x=0, y=0, z=1)),
                (Item.GREEN_CONCRETE, Vec3(x=1, y=0, z=1)),
                (Item.AIR, Vec3(x=-1, y=0, z=0)),
                (Item.AIR, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=1, y=0, z=0)),
                (Item.BLUE_CONCRETE, Vec3(x=-1, y=0, z=-1)),
                (Item.AIR, Vec3(x=0, y=0, z=-1)),
                (Item.YELLOW_CONCRETE, Vec3(x=1, y=0, z=-1))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=1, y=0, z=0)),
                (Item.GREEN_CONCRETE, Vec3(x=2, y=0, z=0)),
                (Item.AIR, Vec3(x=0, y=0, z=1)),
                (Item.AIR, Vec3(x=1, y=0, z=1)),
                (Item.AIR, Vec3(x=2, y=0, z=1)),
                (Item.BLUE_CONCRETE, Vec3(x=0, y=0, z=2)),
                (Item.AIR, Vec3(x=1, y=0, z=2)),
                (Item.YELLOW_CONCRETE, Vec3(x=2, y=0, z=2))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=-2, y=0, z=0)),
                (Item.AIR, Vec3(x=-1, y=0, z=0)),
                (Item.GREEN_CONCRETE, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=-2, y=0, z=1)),
                (Item.AIR, Vec3(x=-1, y=0, z=1)),
                (Item.AIR, Vec3(x=0, y=0, z=1)),
                (Item.BLUE_CONCRETE, Vec3(x=-2, y=0, z=2)),
                (Item.AIR, Vec3(x=-1, y=0, z=2)),
                (Item.YELLOW_CONCRETE, Vec3(x=0, y=0, z=2))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=0, y=0, z=-2)),
                (Item.AIR, Vec3(x=1, y=0, z=-2)),
                (Item.GREEN_CONCRETE, Vec3(x=2, y=0, z=-2)),
                (Item.AIR, Vec3(x=0, y=0, z=-1)),
                (Item.AIR, Vec3(x=1, y=0, z=-1)),
                (Item.AIR, Vec3(x=2, y=0, z=-1)),
                (Item.BLUE_CONCRETE, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=1, y=0, z=0)),
                (Item.YELLOW_CONCRETE, Vec3(x=2, y=0, z=0))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=-2, y=0, z=-2)),
                (Item.AIR, Vec3(x=-1, y=0, z=-2)),
                (Item.GREEN_CONCRETE, Vec3(x=0, y=0, z=-2)),
                (Item.AIR, Vec3(x=-2, y=0, z=-1)),
                (Item.AIR, Vec3(x=-1, y=0, z=-1)),
                (Item.AIR, Vec3(x=0, y=0, z=-1)),
                (Item.BLUE_CONCRETE, Vec3(x=-2, y=0, z=0)),
                (Item.AIR, Vec3(x=-1, y=0, z=0)),
                (Item.YELLOW_CONCRETE, Vec3(x=0, y=0, z=0))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=-1, y=0, z=-1)),
                (Item.AIR, Vec3(x=0, y=0, z=-1)),
                (Item.GREEN_CONCRETE, Vec3(x=1, y=0, z=-1)),
                (Item.AIR, Vec3(x=-1, y=0, z=0)),
                (Item.AIR, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=1, y=0, z=0)),
                (Item.BLUE_CONCRETE, Vec3(x=-1, y=0, z=1)),
                (Item.AIR, Vec3(x=0, y=0, z=1)),
                (Item.YELLOW_CONCRETE, Vec3(x=1, y=0, z=1))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=-1, y=0, z=0)),
                (Item.GREEN_CONCRETE, Vec3(x=-2, y=0, z=0)),
                (Item.AIR, Vec3(x=0, y=-1, z=0)),
                (Item.AIR, Vec3(x=-1, y=-1, z=0)),
                (Item.AIR, Vec3(x=-2, y=-1, z=0)),
                (Item.BLUE_CONCRETE, Vec3(x=0, y=-2, z=0)),
                (Item.AIR, Vec3(x=-1, y=-2, z=0)),
                (Item.YELLOW_CONCRETE, Vec3(x=-2, y=-2, z=0))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=2, y=0, z=0)),
                (Item.AIR, Vec3(x=1, y=0, z=0)),
                (Item.GREEN_CONCRETE, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=2, y=-1, z=0)),
                (Item.AIR, Vec3(x=1, y=-1, z=0)),
                (Item.AIR, Vec3(x=0, y=-1, z=0)),
                (Item.BLUE_CONCRETE, Vec3(x=2, y=-2, z=0)),
                (Item.AIR, Vec3(x=1, y=-2, z=0)),
                (Item.YELLOW_CONCRETE, Vec3(x=0, y=-2, z=0))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=0, y=2, z=0)),
                (Item.AIR, Vec3(x=-1, y=2, z=0)),
                (Item.GREEN_CONCRETE, Vec3(x=-2, y=2, z=0)),
                (Item.AIR, Vec3(x=0, y=1, z=0)),
                (Item.AIR, Vec3(x=-1, y=1, z=0)),
                (Item.AIR, Vec3(x=-2, y=1, z=0)),
                (Item.BLUE_CONCRETE, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=-1, y=0, z=0)),
                (Item.YELLOW_CONCRETE, Vec3(x=-2, y=0, z=0))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=2, y=2, z=0)),
                (Item.AIR, Vec3(x=1, y=2, z=0)),
                (Item.GREEN_CONCRETE, Vec3(x=0, y=2, z=0)),
                (Item.AIR, Vec3(x=2, y=1, z=0)),
                (Item.AIR, Vec3(x=1, y=1, z=0)),
                (Item.AIR, Vec3(x=0, y=1, z=0)),
                (Item.BLUE_CONCRETE, Vec3(x=2, y=0, z=0)),
                (Item.AIR, Vec3(x=1, y=0, z=0)),
                (Item.YELLOW_CONCRETE, Vec3(x=0, y=0, z=0))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=1, y=1, z=0)),
                (Item.AIR, Vec3(x=0, y=1, z=0)),
                (Item.GREEN_CONCRETE, Vec3(x=-1, y=1, z=0)),
                (Item.AIR, Vec3(x=1, y=0, z=0)),
                (Item.AIR, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=-1, y=0, z=0)),
                (Item.BLUE_CONCRETE, Vec3(x=1, y=-1, z=0)),
                (Item.AIR, Vec3(x=0, y=-1, z=0)),
                (Item.YELLOW_CONCRETE, Vec3(x=-1, y=-1, z=0))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=1, y=0, z=0)),
                (Item.GREEN_CONCRETE, Vec3(x=2, y=0, z=0)),
                (Item.AIR, Vec3(x=0, y=-1, z=0)),
                (Item.AIR, Vec3(x=1, y=-1, z=0)),
                (Item.AIR, Vec3(x=2, y=-1, z=0)),
                (Item.BLUE_CONCRETE, Vec3(x=0, y=-2, z=0)),
                (Item.AIR, Vec3(x=1, y=-2, z=0)),
                (Item.YELLOW_CONCRETE, Vec3(x=2, y=-2, z=0))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=-2, y=0, z=0)),
                (Item.AIR, Vec3(x=-1, y=0, z=0)),
                (Item.GREEN_CONCRETE, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=-2, y=-1, z=0)),
                (Item.AIR, Vec3(x=-1, y=-1, z=0)),
                (Item.AIR, Vec3(x=0, y=-1, z=0)),
                (Item.BLUE_CONCRETE, Vec3(x=-2, y=-2, z=0)),
                (Item.AIR, Vec3(x=-1, y=-2, z=0)),
                (Item.YELLOW_CONCRETE, Vec3(x=0, y=-2, z=0))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=0, y=2, z=0)),
                (Item.AIR, Vec3(x=1, y=2, z=0)),
                (Item.GREEN_CONCRETE, Vec3(x=2, y=2, z=0)),
                (Item.AIR, Vec3(x=0, y=1, z=0)),
                (Item.AIR, Vec3(x=1, y=1, z=0)),
                (Item.AIR, Vec3(x=2, y=1, z=0)),
                (Item.BLUE_CONCRETE, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=1, y=0, z=0)),
                (Item.YELLOW_CONCRETE, Vec3(x=2, y=0, z=0))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=-2, y=2, z=0)),
                (Item.AIR, Vec3(x=-1, y=2, z=0)),
                (Item.GREEN_CONCRETE, Vec3(x=0, y=2, z=0)),
                (Item.AIR, Vec3(x=-2, y=1, z=0)),
                (Item.AIR, Vec3(x=-1, y=1, z=0)),
                (Item.AIR, Vec3(x=0, y=1, z=0)),
                (Item.BLUE_CONCRETE, Vec3(x=-2, y=0, z=0)),
                (Item.AIR, Vec3(x=-1, y=0, z=0)),
                (Item.YELLOW_CONCRETE, Vec3(x=0, y=0, z=0))
            ],
            [
                (Item.RED_CONCRETE, Vec3(x=-1, y=1, z=0)),
                (Item.AIR, Vec3(x=0, y=1, z=0)),
                (Item.GREEN_CONCRETE, Vec3(x=1, y=1, z=0)),
                (Item.AIR, Vec3(x=-1, y=0, z=0)),
                (Item.AIR, Vec3(x=0, y=0, z=0)),
                (Item.AIR, Vec3(x=1, y=0, z=0)),
                (Item.BLUE_CONCRETE, Vec3(x=-1, y=-1, z=0)),
                (Item.AIR, Vec3(x=0, y=-1, z=0)),
                (Item.YELLOW_CONCRETE, Vec3(x=1, y=-1, z=0))
            ]
        ]

    def test_offsets(self):
        """Tests multiple offsets."""
        for index, (direction, anchor) in enumerate(product(
                self.directions, self.anchors)):
            self.assertEqual(
                list(offsets(self.valid_profile, direction, anchor)),
                self.results[index])


class TestValidate(TestCase):
    """Test the validate() function."""

    def setUp(self):
        """Sets example profiles."""
        self.valid_profile = [
            [Item.RED_CONCRETE, Item.AIR, Item.GREEN_CONCRETE],
            [Item.AIR, Item.AIR, Item.AIR],
            [Item.BLUE_CONCRETE, Item.AIR, Item.YELLOW_CONCRETE],
        ]
        self.invalid_profile1 = [
            [Item.RED_CONCRETE, Item.AIR, Item.GREEN_CONCRETE],
            [Item.AIR, Item.AIR],
            [Item.BLUE_CONCRETE, Item.AIR, Item.YELLOW_CONCRETE],
        ]
        self.invalid_profile2 = [
            [],
            [Item.AIR, Item.AIR, Item.AIR],
            [Item.BLUE_CONCRETE, Item.AIR, Item.YELLOW_CONCRETE],
        ]
        self.valid_cuboid = [self.valid_profile, self.valid_profile]
        self.invalid_cuboid1 = [self.invalid_profile1, self.invalid_profile1]
        self.valid_row = [Item.RED_CONCRETE, Item.AIR, Item.GREEN_CONCRETE]
        self.invalid_row1 = [Item.RED_CONCRETE, self.valid_profile]

    def test_validate(self):
        """Tests the validation of the profiles."""
        self.assertEqual(validate(self.valid_profile), 2)
        self.assertEqual(validate(self.invalid_profile1), 0)
        self.assertEqual(validate(self.invalid_profile2), 0)
        self.assertEqual(validate(self.invalid_cuboid1), 0)

        self.assertEqual(validate(self.valid_cuboid), 3)
        self.assertEqual(validate(self.invalid_cuboid1), 0)

        self.assertEqual(validate(self.valid_row), 1)
        self.assertEqual(validate(self.invalid_row1), 0)



"""Helper functions."""

from typing import Iterator

import numpy as np
from mcipc.rcon.enumerations import Item

from mcwb.types import (Anchor, Direction, Items, Number, Offsets,
                        Profile, Row, Vec3)

__all__ = ['get_direction', 'normalize', 'offsets', 'validate']


def _get_offset(direction: Vec3, x_start: Number, y_start: Number,
                delta_y: Number, delta_xz: Number) -> Vec3:
    """Returns the offset for a given direction."""

    if direction.north:
        return Vec3(-x_start + delta_xz, y_start - delta_y, 0)

    if direction.south:
        return Vec3(x_start - delta_xz, y_start - delta_y, 0)

    if direction.east:
        return Vec3(0, y_start - delta_y, -x_start + delta_xz)

    if direction.west:
        return Vec3(0, y_start - delta_y, x_start - delta_xz)

    if direction.up:
        return Vec3(-x_start + delta_xz, 0, y_start - delta_y)

    if direction.down:
        return Vec3(-x_start + delta_xz, 0, -y_start + delta_y)

    raise ValueError('Cannot determine offset.')


def get_direction(start: Vec3, end: Vec3) -> Vec3:
    """Checks whether the vetors form a line
    and returns the direction vector.
    """

    if sum(coord1 != coord2 for coord1, coord2 in zip(start, end)) != 1:
        raise ValueError('Not one direction given.')

    return end - start


def normalize(profile: Profile, default: Item = Item.AIR) -> Iterator[Row]:
    """Normalizes a profile."""

    for row in profile:
        yield [default if block is None else Item(block) for block in row]


def offsets(profile: Profile, direction: Vec3, anchor: Anchor) -> Offsets:
    """Yields block offsets dependent on the given direction."""

    height = len(profile)
    width = len(profile[0])
    x_start = y_start = 0

    if isinstance(direction, Direction):
        direction = direction.value

    if anchor in {Anchor.BOTTOM_LEFT, Anchor.BOTTOM_RIGHT}:
        y_start = height - 1
    if anchor in {Anchor.TOP_RIGHT, Anchor.BOTTOM_RIGHT}:
        x_start = width - 1
    elif anchor == Anchor.MIDDLE:
        x_start = int(width / 2)
        y_start = int(height / 2)

    for delta_y, row in enumerate(profile):
        for delta_xz, block in enumerate(row):
            vec3 = _get_offset(direction, x_start, y_start, delta_y, delta_xz)
            yield (block, vec3)


def validate(items: Items):
    """
    Validates a row, profile or cuboid, ensuring consistent dimensions
    """
    ragged = False

    # np ragged array is deprecated but only supplys a warning currently
    with np.warnings.catch_warnings(record=True) as warning:
        np.warnings.simplefilter("always")
        narray = np.array(items)
        if warning and 'ragged nested' in str(warning[-1].message):
            ragged = True

    return int(narray.ndim) if narray.dtype == Item and not ragged else 0


def shift(arr: np.ndarray, vec: Vec3, fill: Item = Item.AIR) -> np.ndarray:
    """ shift a 3d array of Item by vec, discarding the cells that are
        shifted out and filling the new space with Item fill
    """

    # this is the fastest approach in python, see 1d Benchmark at
    # https://stackoverflow.com/questions/30399534/shift-elements-in-a-numpy-array
    result: np.ndarray = np.array(np.full_like(arr, fill), dtype=Item)
    if vec.y > 0:
        result[:, : vec.y, :] = fill
        result[:, vec.y:, :] = arr[:, : -vec.y, :]
    elif vec.y < 0:
        result[:, vec.y:, :] = fill
        result[:, : vec.y, :] = arr[:, -vec.y:, :]
    if vec.x > 0:
        result[: vec.x, :, :] = fill
        result[vec.x:, :, :] = arr[: -vec.x, :, :]
    elif vec.x < 0:
        result[vec.x:, :, :] = fill
        result[: vec.x, :, :] = arr[-vec.x:, :, :]
    if vec.z > 0:
        result[:, :, : vec.z:] = fill
        result[:, :, vec.z:] = arr[:, :, : -vec.z]
    elif vec.z < 0:
        result[:, :, vec.z:] = fill
        result[:, :, : vec.z] = arr[:, :, -vec.z:]

    return result



from typing import cast
from mcwb.types import Planes3d, Vec3
from pathlib import Path
from unittest import TestCase

from mcwb.blocks import Blocks
from mcipc.rcon import Client
from mcwb.itemlists import load_items
from tests.mockclient import MockClient

cubes_dir = Path(__file__).parent / "cubes"


class TestRotation(TestCase):
    """Test the rotate() function."""

    def setUp(self):
        self.client = MockClient()
        self.cube = load_items(cubes_dir / "RGB.cube")
        self.start = Vec3(0, 0, 0)

    def test_XY(self):
        world_cube = Blocks(cast(Client, self.client), self.start, self.cube)
        for rot in range(1, 4):
            world_cube.rotate(Planes3d.XY)
            rotated = load_items(cubes_dir / f"RGBrotateXY{rot}.cube")

            self.assertTrue(self.client.compare(self.start, rotated))

        # 4th rotation returns to orignal state
        world_cube.rotate(Planes3d.XY)
        self.assertTrue(self.client.compare(self.start, self.cube))

    def test_XZ(self):
        world_cube = Blocks(cast(Client, self.client), self.start, self.cube)
        world_cube.rotate(Planes3d.XZ)
        rotated = load_items(cubes_dir / "RGBrotateXZ1.cube")

        self.assertTrue(self.client.compare(self.start, rotated))

    def test_YZ(self):
        world_cube = Blocks(cast(Client, self.client), self.start, self.cube)
        world_cube.rotate(Planes3d.YZ)
        rotated = load_items(cubes_dir / "RGBrotateYZ1.cube")

        self.assertTrue(self.client.compare(self.start, rotated))



"""Builder data types."""

from __future__ import annotations
from enum import Enum
from math import ceil, floor, sqrt
from typing import Iterator, List, NamedTuple, Tuple, Union

from mcipc.rcon.enumerations import Item


__all__ = [
    'Anchor',
    'Anchor3',
    'Cuboid',
    'Direction',
    'Items',
    'Number',
    'Offset',
    'Offsets',
    'Profile',
    'Row',
    'Vec3'
]


Number = Union[float, int]


class Anchor(Enum):
    """Anchor point for tunnels."""

    TOP_LEFT = 'top_left'
    TOP_RIGHT = 'top_right'
    BOTTOM_LEFT = 'bottom_left'
    BOTTOM_RIGHT = 'bottom_right'
    MIDDLE = CENTER = CENTRE = 'middle'


class Anchor3(Enum):
    """Anchor point for cuboids."""

    MIDDLE = CENTER = CENTRE = 'middle'
    BOTTOM_MIDDLE = BOTTOM_CENTER = BOTTOM_CENTRE = 'bottom_middle'
    TOP_MIDDLE = TOP_CENTER = TOP_CENTRE = 'top_middle'

    BOTTOM_SW = 'bottom_sw'
    BOTTOM_NW = 'bottom_nw'
    BOTTOM_NE = 'bottom_ne'
    BOTTOM_SE = 'bottom_se'

    TOP_SW = 'top_sw'
    TOP_NW = 'top_nw'
    TOP_NE = 'top_ne'
    TOP_SE = 'top_se'

    def __str__(self):
        return f'{self.value}'


class Anchor3Face:
    """Groupings of Anchor3 based on which face of the volume they inhabit"""

    TOP = {Anchor3.TOP_SW, Anchor3.TOP_NW, Anchor3.TOP_NE,
           Anchor3.TOP_SE, Anchor3.TOP_MIDDLE}
    BOTTOM = {Anchor3.BOTTOM_SW, Anchor3.BOTTOM_NW,
              Anchor3.BOTTOM_NE, Anchor3.BOTTOM_SE}
    # middle of top or bottom face
    MIDDLE_FACE = CENTER_FACE = CENTRE_FACE = {Anchor3.BOTTOM_MIDDLE,
                                               Anchor3.TOP_MIDDLE}
    NORTH = {Anchor3.TOP_NW, Anchor3.TOP_NE,
             Anchor3.BOTTOM_NW, Anchor3.BOTTOM_NE}
    SOUTH = {Anchor3.TOP_SW, Anchor3.TOP_SE,
             Anchor3.BOTTOM_SW, Anchor3.BOTTOM_SE}
    EAST = {Anchor3.TOP_NE, Anchor3.TOP_SE,
            Anchor3.BOTTOM_NE, Anchor3.BOTTOM_SE}
    WEST = {Anchor3.TOP_SW, Anchor3.TOP_NW,
            Anchor3.BOTTOM_SW, Anchor3.BOTTOM_NW}


class Vec3(NamedTuple):
    """A 3D vector."""

    x: Number = 0
    y: Number = 0
    z: Number = 0

    def __abs__(self):
        return type(self)(abs(self.x), abs(self.y), abs(self.z))

    def __add__(self, other):
        if isinstance(other, Vec3):
            return type(self)(self.x + other.x, self.y + other.y,
                              self.z + other.z)

        return type(self)(self.x + other, self.y + other, self.z + other)

    def __ceil__(self):
        return type(self)(ceil(self.x), ceil(self.y), ceil(self.z))

    def __floor__(self):
        return type(self)(floor(self.x), floor(self.y), floor(self.z))

    def __floordiv__(self, other):
        if isinstance(other, Vec3):
            return type(self)(self.x // other.x, self.y // other.y,
                              self.z // other.z)

        return type(self)(self.x // other, self.y // other, self.z // other)

    def __getitem__(self, index_or_key: Union[int, str]):
        """Returns an item by index or key."""
        if index_or_key in self.keys():
            return getattr(self, index_or_key)

        # Explicitely call to tuple, since
        # super() does not work in NamedTuples.
        return tuple.__getitem__(self, index_or_key)

    def __mul__(self, other):
        if isinstance(other, Vec3):
            return self.__matmul__(other)

        return type(self)(self.x * other, self.y * other, self.z * other)

    def __matmul__(self, other: Vec3):
        return type(self)(self.x * other.x, self.y * other.y, self.z * other.z)

    def __neg__(self):
        return type(self)(-self.x, -self.y, -self.z)

    def __pos__(self):
        return self     # No operation necessary.

    def __round__(self, ndigits: int = 0):
        return type(self)(round(self.x, ndigits), round(self.y, ndigits),
                          round(self.z, ndigits))

    def __sub__(self, other):
        if isinstance(other, Vec3):
            return type(self)(self.x - other.x, self.y - other.y,
                              self.z - other.z)

        return type(self)(self.x - other, self.y - other, self.z - other)

    def __truediv__(self, other):
        if isinstance(other, Vec3):
            return type(self)(self.x / other.x, self.y / other.y,
                              self.z / other.z)

        return type(self)(self.x / other, self.y / other, self.z / other)

    def keys(self):
        """Returns the keys."""
        return self._fields

    def with_ints(self):
        """Coerce all coordinates to int."""
        # floor is consistent with minecraft's handling of negative coords
        return type(self)(floor(self.x), floor(self.y), floor(self.z))

    @property
    def volume(self):
        """Returns the volume of the space described by (0,0,0) -> self."""
        return self.dx * self.dy * self.dz

    @property
    def is_direction(self):
        """Determines whether this is a direction vector."""
        return sum(coord != 0 for coord in self) == 1

    @property
    def dx(self):   # pylint: disable=C0103
        """Returns the absolute x value."""
        return abs(self.x)

    @property
    def dy(self):   # pylint: disable=C0103
        """Returns the absolute y value."""
        return abs(self.y)

    @property
    def dz(self):   # pylint: disable=C0103
        """Returns the absolute z value."""
        return abs(self.z)

    @property
    def length(self):
        """Returns the length of the spatial diagonal."""
        return sqrt(pow(self.dx, 2) + pow(self.dy, 2) + pow(self.dz, 2))

    @property
    def west(self):
        """Checks whether the vector points west."""
        return self.x < 0

    @property
    def east(self):
        """Checks whether the vector points east."""
        return self.x > 0

    @property
    def north(self):
        """Checks whether the vector points north."""
        return self.z < 0

    @property
    def south(self):
        """Checks whether the vector points south."""
        return self.z > 0

    @property
    def down(self):
        """Checks whether the vector points down."""
        return self.y < 0

    @property
    def up(self):   # pylint: disable=C0103
        """Checks whether the vector points up."""
        return self.y > 0


class Direction(Enum):
    """Available directions."""

    EAST = Vec3(+1, 0, 0)
    WEST = Vec3(-1, 0, 0)
    UP = Vec3(0, +1, 0)
    DOWN = Vec3(0, -1, 0)
    SOUTH = Vec3(0, 0, +1)
    NORTH = Vec3(0, 0, -1)


class Planes3d(Enum):
    """
    Represents all 3 planes in a 3d space as a pair of
    dimension indeces as used in np.array: 0=X 1=Y 2=Z """
    XY = (0, 1)
    XZ = (0, 2)
    YZ = (1, 2)


Row = List[Item]
Profile = List[Row]
Cuboid = List[Profile]

Items = Union[Cuboid, Profile, Row]

Offset = Tuple[Item, Vec3]
Offsets = Iterator[Offset]



"""Tests for mcwb.itemlists unit."""

import shutil
import tempfile
from pathlib import Path
from unittest import TestCase

from mcipc.rcon.item import Item
from mcwb.itemlists import load_items, save_items


class TestItemLists(TestCase):  # pylint: disable=R0902
    """Tests saving and loading lists of Items"""

    def setUp(self):
        """Sets example shapes and a temporary directory."""
        self.test_dir = Path(tempfile.mkdtemp())

        self.blue_row = [
            Item.BLUE_CONCRETE,
            Item.BLUE_CONCRETE,
            Item.BLUE_CONCRETE
        ]
        self.red_row = [
            Item.RED_CONCRETE,
            Item.RED_CONCRETE,
            Item.RED_CONCRETE
        ]
        self.hollow_row = [
            Item.RED_CONCRETE,
            Item.AIR,
            Item.RED_CONCRETE
        ]
        self.top_profile = [
            self.blue_row,
            self.red_row,
            self.red_row
        ]
        self.middle_profile = [
            self.blue_row,
            self.hollow_row,
            self.red_row
        ]
        self.bottom_profile = [
            self.blue_row,
            self.blue_row,
            self.red_row
        ]
        self.cube = [
            self.top_profile,
            self.middle_profile,
            self.bottom_profile
        ]

    def tearDown(self):
        """Remove the directory after the test."""
        shutil.rmtree(str(self.test_dir))

    def test_save_row(self):
        """Tests the saving of rows."""
        path = self.test_dir / "row_test.json"
        save_items(self.blue_row, path)
        items = load_items(path, 1)

        self.assertEqual(items, self.blue_row)

    def test_save_profile(self):
        """Tests the saving of profiles."""
        path = self.test_dir / "row_profile.json"
        save_items(self.middle_profile, path)
        items = load_items(path, 2)

        self.assertEqual(items, self.middle_profile)

    def test_save_cuboid(self):
        """Tests the saving of cuboids."""
        path = self.test_dir / "row_cuboid.json"
        save_items(self.cube, path)
        items = load_items(path, 3)

        self.assertEqual(items, self.cube)

        # validate failed dimension check
        with self.assertRaises(ValueError):
            load_items(path, 2)

        # validate the 'any dimensions' call to load_items
        items = load_items(path)
        self.assertEqual(items, self.cube)



"""Methods to build in the world."""

from mcwb.api import mktunnel
from mcwb.types import (
    Anchor, Anchor3, Anchor3Face, Cuboid, Direction, Items, Profile, Row, Vec3
)
from mcwb.volume import Volume


__all__ = ['Anchor', 'Anchor3', 'Anchor3Face', 'Cuboid', 'Direction', 'Items',
           'Profile', 'Row', 'Vec3', 'Volume',
           'mktunnel'
           ]



""" render and transform cuboids of blocks in minecraft space """
from typing import Any, Union

import numpy as np
from mcipc.rcon.enumerations import Item
from mcipc.rcon.je import Client
from mcwb import Anchor3, Vec3, Volume
from mcwb.types import Cuboid, Items, Planes3d

from mcwb.functions import shift


class Blocks:
    """
    Represents a cubiod of arbitrary blocks in a minecraft world with functions
    for transforming and rendering those blocks in the world
    """

    def __init__(
        self,
        client: Client,
        position: Vec3,
        cube: Union[Items, np.ndarray],
        anchor: Anchor3 = Anchor3.BOTTOM_NW,
        render: bool = True,
    ) -> None:
        self._client = client
        self.anchor = anchor

        if isinstance(cube, np.ndarray):
            self.ncube = cube
        else:
            self.ncube = np.array(cube, dtype=Item)

        if self.ncube.ndim != 3:
            raise ValueError("invalid cube specificaton")

        self.volume = Volume.from_anchor(
            position, Vec3(*self.ncube.shape), self.anchor)
        self._solid: Any = self.ncube != Item.AIR

        if render:
            self._render()

    def _render(self) -> None:
        """ render the blocks into Minecraft """
        for idx, block in np.ndenumerate(self.ncube):
            if self._solid[idx]:
                self._client.setblock(self.volume.start + Vec3(*idx), block)

    def _unrender(self, vector: Vec3, old_start: Vec3) -> None:
        """ clear away exposed blocks from the previous move """
        moved = shift(self.ncube, vector * 1)

        mask: Any = (moved == Item.AIR) & (self.ncube != Item.AIR)

        for idx, block in np.ndenumerate(self.ncube):
            if mask[idx]:
                self._client.setblock(old_start + Vec3(*idx), Item.AIR.value)

    def rotate(self, plane: Planes3d, steps: int = 1, clear=True) -> None:
        """ rotate the blocks in place """
        self.ncube = np.rot90(self.ncube, k=steps, axes=plane.value)

        if clear:  # TODO implement unrender for rotated blocks (challenging?)
            self.volume.fill(self._client)

        self._solid = self.ncube != Item.AIR
        self.volume = Volume.from_anchor(
            self.volume.position, Vec3(*self.ncube.shape), self.anchor
        )

        self._render()

    def move(self, vector: Vec3, clear: bool = True) -> None:
        """ moves the cubiod by vector and redraws it """
        old_start = self.volume.start
        self.volume = Volume.from_anchor(
            self.volume.position + vector, Vec3(*self.ncube.shape), self.anchor
        )

        self._render()

        if clear:
            self._unrender(vector, old_start)

    def move_to(self, position: Vec3, clear: bool = True) -> None:
        """ moves the cubiod to position and redraws it """
        old_volume = self.volume
        self.volume = Volume.from_anchor(
            position, Vec3(*self.ncube.shape), self.anchor)

        self._render()

        if clear:
            old_volume.fill(self._client)

    def to_cuboid(self) -> Cuboid:
        """ return the blocks' contents as a Cuboid """
        return self.ncube.tolist()



"""Representation of 3D spaces."""

from __future__ import annotations
from mcipc.rcon.enumerations import Item
from mcipc.rcon.je import Client

from mcwb import Anchor, Anchor3, Anchor3Face, Direction, Vec3, mktunnel

__all__ = ['Volume']

MAX_MINECRAFT_FILL_COMMAND = 32768


class Volume:
    """
    Describes a 3d space in a Minecraft world using one of:
      - a starting point and size. The starting point can be any vertex or
        the middle of one of the horizontal faces, using the cardinal
        terminology defined in Anchor3.
      - two opposite corners
    """

    def __init__(self):
        self.start = Vec3(0, 0, 0)  # lowest coordinate in the volume
        self.end = Vec3(0, 0, 0)  # highest coordinate in the volume
        self.size = Vec3(0, 0, 0)  # the dimensions of the volume

        # NOTE: position == start when self.anchor == Anchor3.BOTTOM_NW
        self.anchor = Anchor3.BOTTOM_NW  # the anchor point for position
        self.position = Vec3(0, 0, 0)  # the positon of the anchor point

    @classmethod
    def from_anchor(cls, position: Vec3, size: Vec3, anchor: Anchor3) -> Volume:
        """ a factory function to create a Volume using anchor and size """
        volume = Volume()
        volume.position = Vec3(*position)
        volume.size = Vec3(*size)
        volume.anchor = anchor

        offset = Vec3(0, 0, 0)
        if anchor in Anchor3Face.TOP:
            offset += Vec3(0, 1 - size.y, 0)
        if anchor in Anchor3Face.SOUTH:
            offset += Vec3(0, 0, 1 - size.z)
        if anchor in Anchor3Face.EAST:
            offset += Vec3(1 - size.x, 0, 0)
        if anchor in Anchor3Face.MIDDLE_FACE:  # middle of top or bottom
            offset -= Vec3(int(size.x / 2), 0, int(size.z / 2))
        elif anchor is Anchor3.MIDDLE:
            offset = Vec3(-1, -1, -1) * (size / 2).with_ints()

        volume.start = position + offset
        volume.end = volume.start + (volume.size - 1)

        return volume

    @classmethod
    def from_corners(cls, start: Vec3, end: Vec3) -> Volume:
        """ a factory function to define Volume using opposite corners """
        volume = Volume()
        start = Vec3(*start)
        end = Vec3(*end)

        # normalize start and end so all start coords are minima
        volume.position = volume.start = Vec3(
            min(start.x, end.x), min(
                start.y, end.y), min(start.z, end.z)
        )
        volume.end = Vec3(
            max(start.x, end.x), max(
                start.y, end.y), max(start.z, end.z)
        )
        volume.size = (volume.end - volume.start) + 1

        return volume

    def inside(
        self, position: Vec3, xtol: int = 0, ytol: int = 0, ztol: int = 0
    ) -> bool:
        """ determine if position is within the Volume """
        return (
            self.start.x - xtol <= position.x <= self.end.x + xtol
            and self.start.y - ytol <= position.y <= self.end.y + ytol
            and self.start.z - ztol <= position.z <= self.end.z + ztol
        )

    def move(self, distance: Vec3) -> None:
        """ move the volume's location in space by distance """
        self.start += distance
        self.end += distance
        self.position += distance

    def move_to(self, position: Vec3) -> None:
        """ move the volume's location in space to position """
        self.start += position - self.position
        self.end += position - self.position
        self.position = position

    def fill(self, client: Client, block: Item = Item.AIR):
        """ Fill the Volume with a single block type, supports large volumes """
        if self.size.volume < MAX_MINECRAFT_FILL_COMMAND:
            client.fill(self.start, self.end, block.value)
        else:
            profile = [[block] * int(self.size.x)] * int(self.size.y)
            mktunnel(
                client,
                profile,
                self.start,
                direction=Direction.UP,
                anchor=Anchor.BOTTOM_LEFT,
                length=int(self.size.dz),
            )

    def walls(
        self,
        client: Client,
        block: Item,
        *,
        thickness: int = 1,
        top: bool = True,
        bottom: bool = True,
        north: bool = True,
        south: bool = True,
        east: bool = True,
        west: bool = True,
    ) -> None:
        """ renders walls at the faces of the volume """
        t = thickness - 1
        b = block.value
        if north:
            client.fill(self.start, Vec3(
                self.end.x, self.end.y, self.start.z + t), b)
        if south:
            client.fill(self.end, Vec3(
                self.start.x, self.start.y, self.end.z - t), b)
        if west:
            client.fill(self.start, Vec3(
                self.start.x + t, self.end.y, self.end.z), b)
        if east:
            client.fill(self.end, Vec3(self.end.x - t,
                                       self.start.y, self.start.z), b)
        if top:
            client.fill(self.end, Vec3(
                self.start.x, self.end.y - t, self.start.z), b)
        if bottom:
            client.fill(self.start, Vec3(
                self.end.x, self.start.y + 1, self.end.z), b)



#! /usr/bin/env python
"""Installation script."""

from setuptools import setup

setup(
    name='mcwb',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    author='Richard Neumann',
    author_email='mail@richard-neumann.de',
    python_requires='>=3.8',
    packages=['mcwb'],
    install_requires=['mcipc'],
    url='https://github.com/conqp/mcwb',
    license='GPLv3',
    description='A Minecraft world builder library.',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    keywords='minecraft world builder python rcon'
)



from itertools import product
from unittest import TestCase

from mcwb import Volume, Anchor3, Vec3


class TestVolume(TestCase):
    """Tests the Volume constructor."""

    def test_anchors_odd_size(self):
        size = Vec3(3, 3, 3)
        position = Vec3(0, 0, 0)

        v = Volume.from_anchor(position, size, anchor=Anchor3.MIDDLE)
        self.assertTrue(v.position == position)
        self.assertTrue(v.start == Vec3(-1, -1, -1))
        self.assertTrue(v.end == Vec3(1, 1, 1))
        self.assertTrue(v.size == size)

        v = Volume.from_anchor(position, size, anchor=Anchor3.BOTTOM_NW)
        self.assertTrue(v.position == position)
        self.assertTrue(v.start == Vec3(0, 0, 0))
        self.assertTrue(v.end == Vec3(2, 2, 2))
        self.assertTrue(v.size == size)

        v = Volume.from_anchor(position, size, anchor=Anchor3.BOTTOM_NE)
        self.assertTrue(v.position == position)
        self.assertTrue(v.start == Vec3(-2, 0, 0))
        self.assertTrue(v.end == Vec3(0, 2, 2))
        self.assertTrue(v.size == size)

        v = Volume.from_anchor(position, size, anchor=Anchor3.BOTTOM_SE)
        self.assertTrue(v.position == position)
        self.assertTrue(v.start == Vec3(-2, 0, -2))
        self.assertTrue(v.end == Vec3(0, 2, 0))
        self.assertTrue(v.size == size)

        v = Volume.from_anchor(position, size, anchor=Anchor3.BOTTOM_SW)
        self.assertTrue(v.position == position)
        self.assertTrue(v.start == Vec3(0, 0, -2))
        self.assertTrue(v.end == Vec3(2, 2, 0))
        self.assertTrue(v.size == size)

        v = Volume.from_anchor(position, size, anchor=Anchor3.BOTTOM_MIDDLE)
        self.assertTrue(v.position == position)
        self.assertTrue(v.start == Vec3(-1, 0, -1))
        self.assertTrue(v.end == Vec3(1, 2, 1))
        self.assertTrue(v.size == size)

        v = Volume.from_anchor(position, size, anchor=Anchor3.TOP_NW)
        self.assertTrue(v.position == position)
        self.assertTrue(v.start == Vec3(0, -2, 0))
        self.assertTrue(v.end == Vec3(2, 0, 2))
        self.assertTrue(v.size == size)

        v = Volume.from_anchor(position, size, anchor=Anchor3.TOP_NE)
        self.assertTrue(v.position == position)
        self.assertTrue(v.start == Vec3(-2, -2, 0))
        self.assertTrue(v.end == Vec3(0, 0, 2))
        self.assertTrue(v.size == size)

        v = Volume.from_anchor(position, size, anchor=Anchor3.TOP_SE)
        self.assertTrue(v.position == position)
        self.assertTrue(v.start == Vec3(-2, -2, -2))
        self.assertTrue(v.end == Vec3(0, 0, 0))
        self.assertTrue(v.size == size)

        v = Volume.from_anchor(position, size, anchor=Anchor3.TOP_SW)
        self.assertTrue(v.position == position)
        self.assertTrue(v.start == Vec3(0, -2, -2))
        self.assertTrue(v.end == Vec3(2, 0, 0))
        self.assertTrue(v.size == size)

        v = Volume.from_anchor(position, size, anchor=Anchor3.TOP_MIDDLE)
        self.assertTrue(v.position == position)
        self.assertTrue(v.start == Vec3(-1, -2, -1))
        self.assertTrue(v.end == Vec3(1, 0, 1))
        self.assertTrue(v.size == size)

    def test_anchors_even_size(self):
        size = Vec3(4, 4, 4)
        position = Vec3(0, 0, 0)

        v = Volume.from_anchor(position, size, anchor=Anchor3.MIDDLE)
        self.assertTrue(v.position == position)
        self.assertTrue(v.start == Vec3(-2, -2, -2))
        self.assertTrue(v.end == Vec3(1, 1, 1))
        self.assertTrue(v.size == size)

        v = Volume.from_anchor(position, size, anchor=Anchor3.BOTTOM_NW)
        self.assertTrue(v.position == position)
        self.assertTrue(v.start == Vec3(0, 0, 0))
        self.assertTrue(v.end == Vec3(3, 3, 3))
        self.assertTrue(v.size == size)

        v = Volume.from_anchor(position, size, anchor=Anchor3.BOTTOM_NE)
        self.assertTrue(v.position == position)
        self.assertTrue(v.start == Vec3(-3, 0, 0))
        self.assertTrue(v.end == Vec3(0, 3, 3))
        self.assertTrue(v.size == size)

    def test_bounds_odd_size(self):
        size = Vec3(3, 3, 3)
        position = Vec3(-1, -1, -1)

        for x, y, z in product([1, -1], [1, -1], [1, -1]):
            start = Vec3(x, y, z)
            opposite = Vec3(-x, -y, -z)
            print(start, opposite)
            v = Volume.from_corners(start, end=opposite)
            print(v.start, v.end, v.position, v.size)
            self.assertTrue(v.position == position)
            self.assertTrue(v.size == size)



"""Exposed API functions."""

from mcipc.rcon.client import Client
from mcipc.rcon.enumerations import FillMode, Item

from mcwb.functions import get_direction
from mcwb.functions import offsets
from mcwb.functions import normalize
from mcwb.functions import validate
from mcwb.types import Anchor, Direction, Profile, Vec3


__all__ = ['mktunnel']


def mktunnel(client: Client, profile: Profile, start: Vec3, *,
             end: Vec3 = None, direction: Direction = None, length: int = 1,
             anchor: Anchor = Anchor.BOTTOM_RIGHT, default: Item = Item.AIR,
             mode: FillMode = None, filter: str = None):
    """Creates a tunnel with the given profile."""

    start = Vec3(*start)    # Ensure Vec3 object.

    if validate(profile) != 2:
        raise ValueError('Invalid matrix.')

    if end is None:
        end = start + direction.value * length
    else:
        end = Vec3(*end)    # Ensure Vec3 object.

    calculated_direction = get_direction(start, end)
    profile = list(normalize(profile, default=default))

    for block, offset in offsets(profile, calculated_direction, anchor):
        client.fill(
            start + offset, end + offset, block, mode=mode, filter=filter
        )



"""
Functions for manipulating lists of Items
    1d = Row
    2d = Profile
    3d = Cuboid
"""

import codecs
import json
import re
from pathlib import Path
from typing import Union

import numpy as np
from mcipc.rcon.enumerations import Item
from mcipc.rcon.je import Client

from mcwb import Cuboid, Vec3, Volume, Items
from mcwb.functions import validate


ITEM_KEY = "__Item__"


def save_items(items: Items, filename: Path) -> None:
    """ save a Profile, Cuboid or Row to a json file """

    def json_item(item: Item):
        return {ITEM_KEY: item.value}

    if validate(items) == 0:
        raise ValueError("items is not a valid Row, Profile or Cuboid")

    json.dump(
        items,
        codecs.open(str(filename), "w", encoding="utf-8"),
        separators=(",", ":"),
        sort_keys=True,
        indent=4,
        default=json_item,
    )


def load_items(filename: Union[Path, str], dimensions: int = None) -> Items:
    """ load a JSON file of Items - returns a Cuboid, Profile or Row"""

    def as_item(dct: dict):
        if ITEM_KEY in dct:
            return Item(dct[ITEM_KEY])

        return dct

    result = json.load(
        codecs.open(str(filename), "r", encoding="utf-8"), object_hook=as_item
    )
    valid_dims = validate(result)

    if valid_dims == 0:
        raise ValueError("file {filename} contains invalid JSON")

    if dimensions is not None and valid_dims != dimensions:
        raise ValueError(
            "file {filename} contains {valid_dims} dimension Items"
            "but {dimensions} dimensions was requested"
            )

    return result


dump = Vec3(0, 0, 0)
extract_item = re.compile(r".*minecraft\:(?:blocks\/)?(.+)$")


def grab(client: Client, vol: Volume) -> Cuboid:
    """ copy blocks from a Volume in the minecraft world into a cuboid of Item """
    ncube = np.ndarray(vol.size, dtype=Item)

    for idx, _ in np.ndenumerate(ncube):
        # currently the only way to test for a block is to use the loot spawn
        # command, this creates an entity at 0, 0, 0 that falls into the void
        output = client.loot.spawn(dump).mine(vol.start + Vec3(*idx))
        match = extract_item.search(output)
        if not match:
            raise ValueError(f"unexpected response from loot spawn: {output}")
        name = match.group(1)
        if name == "empty":
            name = "air"
        ncube[idx] = Item(name)

    result = ncube.tolist()
    return result
